package com.sun.xml.internal.bind.v2.runtime;

import com.sun.xml.internal.bind.Util;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.UnmarshallerImpl;
import java.lang.reflect.Method;
import java.util.logging.Level;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.Loader;
import javax.xml.stream.XMLStreamException;
import java.io.IOException;
import org.xml.sax.SAXException;
import java.lang.reflect.InvocationTargetException;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.UnmarshallingContext;
import com.sun.istack.internal.NotNull;
import java.util.Collections;
import java.util.Arrays;
import java.util.Collection;
import javax.xml.namespace.QName;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeTypeInfo;
import java.util.logging.Logger;

public abstract class JaxBeanInfo<BeanT>
{
    protected boolean isNilIncluded;
    protected short flag;
    private static final short FLAG_IS_ELEMENT = 1;
    private static final short FLAG_IS_IMMUTABLE = 2;
    private static final short FLAG_HAS_ELEMENT_ONLY_CONTENTMODEL = 4;
    private static final short FLAG_HAS_BEFORE_UNMARSHAL_METHOD = 8;
    private static final short FLAG_HAS_AFTER_UNMARSHAL_METHOD = 16;
    private static final short FLAG_HAS_BEFORE_MARSHAL_METHOD = 32;
    private static final short FLAG_HAS_AFTER_MARSHAL_METHOD = 64;
    private static final short FLAG_HAS_LIFECYCLE_EVENTS = 128;
    private LifecycleMethods lcm;
    public final Class<BeanT> jaxbType;
    private final Object typeName;
    private static final Class[] unmarshalEventParams;
    private static Class[] marshalEventParams;
    private static final Logger logger;
    
    protected JaxBeanInfo(JAXBContextImpl grammar, RuntimeTypeInfo rti, Class<BeanT> jaxbType, QName[] typeNames, boolean isElement, boolean isImmutable, boolean hasLifecycleEvents) {
        this(grammar, rti, jaxbType, typeNames, isElement, isImmutable, hasLifecycleEvents);
    }
    
    protected JaxBeanInfo(JAXBContextImpl grammar, RuntimeTypeInfo rti, Class<BeanT> jaxbType, QName typeName, boolean isElement, boolean isImmutable, boolean hasLifecycleEvents) {
        this(grammar, rti, jaxbType, typeName, isElement, isImmutable, hasLifecycleEvents);
    }
    
    protected JaxBeanInfo(JAXBContextImpl grammar, RuntimeTypeInfo rti, Class<BeanT> jaxbType, boolean isElement, boolean isImmutable, boolean hasLifecycleEvents) {
        this(grammar, rti, jaxbType, null, isElement, isImmutable, hasLifecycleEvents);
    }
    
    private JaxBeanInfo(JAXBContextImpl grammar, RuntimeTypeInfo rti, Class<BeanT> jaxbType, Object typeName, boolean isElement, boolean isImmutable, boolean hasLifecycleEvents) {
        this.isNilIncluded = false;
        this.lcm = null;
        grammar.beanInfos.put(rti, this);
        this.jaxbType = jaxbType;
        this.typeName = typeName;
        this.flag = (short)((isElement ? 1 : 0) | (isImmutable ? 2 : 0) | (hasLifecycleEvents ? 128 : 0));
    }
    
    public final boolean hasBeforeUnmarshalMethod() {
        return (this.flag & 0x8) != 0x0;
    }
    
    public final boolean hasAfterUnmarshalMethod() {
        return (this.flag & 0x10) != 0x0;
    }
    
    public final boolean hasBeforeMarshalMethod() {
        return (this.flag & 0x20) != 0x0;
    }
    
    public final boolean hasAfterMarshalMethod() {
        return (this.flag & 0x40) != 0x0;
    }
    
    public final boolean isElement() {
        return (this.flag & 0x1) != 0x0;
    }
    
    public final boolean isImmutable() {
        return (this.flag & 0x2) != 0x0;
    }
    
    public final boolean hasElementOnlyContentModel() {
        return (this.flag & 0x4) != 0x0;
    }
    
    protected final void hasElementOnlyContentModel(boolean value) {
        if (value) {
            this.flag |= 0x4;
        }
        else {
            this.flag &= 0xFFFFFFFB;
        }
    }
    
    public boolean isNilIncluded() {
        return this.isNilIncluded;
    }
    
    public boolean lookForLifecycleMethods() {
        return (this.flag & 0x80) != 0x0;
    }
    
    public abstract String getElementNamespaceURI(BeanT p0);
    
    public abstract String getElementLocalName(BeanT p0);
    
    public Collection<QName> getTypeNames() {
        if (this.typeName == null) {
            return Collections.emptyList();
        }
        if ((this.typeName instanceof QName)) {
            return Collections.singletonList((QName)this.typeName);
        }
        return Arrays.asList((QName[])(QName[])this.typeName);
    }
    
    public QName getTypeName(@NotNull BeanT instance) {
        if (this.typeName == null) {
            return null;
        }
        if ((this.typeName instanceof QName)) {
            return (QName)this.typeName;
        }
        return ((QName[])(QName[])this.typeName)[0];
    }
    
    public abstract BeanT createInstance(UnmarshallingContext p0) throws IllegalAccessException, InvocationTargetException, InstantiationException, SAXException;
    
    public abstract boolean reset(BeanT p0, UnmarshallingContext p1) throws SAXException;
    
    public abstract String getId(BeanT p0, XMLSerializer p1) throws SAXException;
    
    public abstract void serializeBody(BeanT p0, XMLSerializer p1) throws SAXException, IOException, XMLStreamException;
    
    public abstract void serializeAttributes(BeanT p0, XMLSerializer p1) throws SAXException, IOException, XMLStreamException;
    
    public abstract void serializeRoot(BeanT p0, XMLSerializer p1) throws SAXException, IOException, XMLStreamException;
    
    public abstract void serializeURIs(BeanT p0, XMLSerializer p1) throws SAXException;
    
    public abstract Loader getLoader(JAXBContextImpl p0, boolean p1);
    
    public abstract Transducer<BeanT> getTransducer();
    
    protected void link(JAXBContextImpl grammar) {
    }
    
    public void wrapUp() {
    }
    
    protected final void setLifecycleFlags() {
        try {
            Class<BeanT> jt = this.jaxbType;
            if (this.lcm == null) {
                this.lcm = new LifecycleMethods();
            }
            while (jt != null) {
                for (Method m : jt.getDeclaredMethods()) {
                    String name = m.getName();
                    if (this.lcm.beforeUnmarshal == null && name.equals("beforeUnmarshal") && this.match(m, JaxBeanInfo.unmarshalEventParams)) {
                        this.cacheLifecycleMethod(m, (short)8);
                    }
                    if (this.lcm.afterUnmarshal == null && name.equals("afterUnmarshal") && this.match(m, JaxBeanInfo.unmarshalEventParams)) {
                        this.cacheLifecycleMethod(m, (short)16);
                    }
                    if (this.lcm.beforeMarshal == null && name.equals("beforeMarshal") && this.match(m, JaxBeanInfo.marshalEventParams)) {
                        this.cacheLifecycleMethod(m, (short)32);
                    }
                    if (this.lcm.afterMarshal == null && name.equals("afterMarshal") && this.match(m, JaxBeanInfo.marshalEventParams)) {
                        this.cacheLifecycleMethod(m, (short)64);
                    }
                }
                jt = jt.getSuperclass();
            }
        }
        catch (SecurityException e) {
            JaxBeanInfo.logger.log(Level.WARNING, Messages.UNABLE_TO_DISCOVER_EVENTHANDLER.format(new Object[] { this.jaxbType.getName(), e }));
        }
    }
    
    private boolean match(Method m, Class[] params) {
        return Arrays.equals(m.getParameterTypes(), params);
    }
    
    private void cacheLifecycleMethod(Method m, short lifecycleFlag) {
        if (this.lcm == null) {
            this.lcm = new LifecycleMethods();
        }
        m.setAccessible(true);
        this.flag |= lifecycleFlag;
        switch (lifecycleFlag) {
            case 8: {
                this.lcm.beforeUnmarshal = m;
                break;
            }
            case 16: {
                this.lcm.afterUnmarshal = m;
                break;
            }
            case 32: {
                this.lcm.beforeMarshal = m;
                break;
            }
            case 64: {
                this.lcm.afterMarshal = m;
                break;
            }
        }
    }
    
    public final LifecycleMethods getLifecycleMethods() {
        return this.lcm;
    }
    
    public final void invokeBeforeUnmarshalMethod(UnmarshallerImpl unm, Object child, Object parent) throws SAXException {
        Method m = this.getLifecycleMethods().beforeUnmarshal;
        this.invokeUnmarshallCallback(m, child, unm, parent);
    }
    
    public final void invokeAfterUnmarshalMethod(UnmarshallerImpl unm, Object child, Object parent) throws SAXException {
        Method m = this.getLifecycleMethods().afterUnmarshal;
        this.invokeUnmarshallCallback(m, child, unm, parent);
    }
    
    private void invokeUnmarshallCallback(Method m, Object child, UnmarshallerImpl unm, Object parent) throws SAXException {
        try {
            m.invoke(child, new Object[] { unm, parent });
        }
        catch (IllegalAccessException e) {
            UnmarshallingContext.getInstance().handleError(e, false);
        }
        catch (InvocationTargetException e2) {
            UnmarshallingContext.getInstance().handleError(e2, false);
        }
    }
    
    static {
        unmarshalEventParams = new Class[] { Unmarshaller.class, Object.class };
        JaxBeanInfo.marshalEventParams = new Class[] { Marshaller.class };
        logger = Util.getClassLogger();
    }
}
